
; 7.5 bit sample player for C64
; (C) 2001 Levente Hrsfalvi

; Ideas obtained from:
;
; - Steve Judd  (C= Hacking 18)
; - Dag Lem (reSID emulator engine)

; Runs on a stock C64 (PAL or NTSC), and plays an unsigned 8-bit sample
; using PWM.

; Sampling rate: 68 clock cycles
; 17734470 / 18 / 68 = 14488.946078 Hz for PAL
; 14318180 / 14 / 68 = 15040.105042 Hz for NTSC

; Table Driven Assembler + Comlink
; (probably easy to compile using another 6502 cross assemblers too)

; This routine is 'free software', I just don't bother quoting GPL


#ifndef LO
#define LO(x) (x&$ff)
#define HI(x) (x>>8)
#endif

         *= $0900

        *= *-2
        .word *+2                       ;Load address for the binary

        sei
        lda #$0b
        sta $d011                       ;blank screen
        lda #$35
        sta $01                         ;ROMs off

        lda #$0f                        ;Volume + filter
        sta $d418
;        lda #$07                        ;Lowpass filter (but it sounded so
;        sta $d417                       ;bad on my 6581R4 that I commented
;        lda #$a0                        ;it out
;        sta $d416

        lda #$00                        ;Set SID freq registers
        sta $d400
        sta $d407
        sta $d40e
        lda #$40                        ;$4000 --> the wave accu is
        sta $d401                       ;incremented by $004 each CPU clock
        sta $d408                       ;cycles
        sta $d40f
        lda #$f0                        ;maximum sustain level for all chns
        sta $d406
        sta $d40d
        sta $d414

        lda #LO($0a00)                  ;start address of sample in memory
        sta ptr
        lda #HI($0a00)
        sta ptr+1
        lda #$cd                        ;end address high. Zpg ($fb) is used
        sta $fb                         ;because of timing reasons

        ldx #$49                        ;pre-load values for the SID command
        ldy #$41                        ;registers...

ptr     =*+1
loop    lda $face    ;4                 ;Load next sample...
        sta $d402    ;8                 ;put to the sample width reg LO
        stx $d404    ;12                ;RESET SID voice
        sty $d404    ;16                ;...then start wave counter
        adc #$01     ;18                ;Increase sample value
        sta $d409    ;22                ;(presumably, there'll never be a
        stx $d40b    ;26                ;carry), then go on for next SID
        sty $d40b    ;30                ;channels.
        adc #$01     ;32
        sta $d410    ;36
        stx $d412    ;40
        sty $d412    ;44

        inc ptr      ;50                ;Increase sample pointer
        bne loop1    ;52  (when jumps, 53)
        inc ptr+1    ;58
        lda ptr+1    ;62
        cmp $fb      ;65
        bne loop     ;68

        lda #LO($0a00)                  ;Reload it (timing is broken, but no
        sta ptr                         ;matter since it only happens per one
        lda #HI($0a00)                  ;full sample replay)
        sta ptr+1
        jmp loop

loop1   bit $face    ;57                ;timing...
        bit $face    ;61
        bit $face    ;65
        jmp loop     ;68                ;Then loop.


i_sample =$0a00                         ;Just to include the sample binary
                                        ;(TASM + Comlink feature)
        .export i_sample

.end
